from ...messages.base import MessageDataModel, MessageMetaModel,MessageModel

from typing import Optional
from ...messages.types import ACTIVATE_POD


class MetaModel(MessageMetaModel):
    type: str = ACTIVATE_POD

class DataModel(MessageDataModel):
    pod: Optional[int]


class ActivatePodMessage(MessageModel):
    data: DataModel
    meta: MetaModel




from bergen.queries.delayed.node import DETAIL_NODE_FR
from bergen.query import DelayedGQL


TEMPLATE_GET_QUERY = DelayedGQL("""
query Template($id: ID,){
  template(id: $id){
    id
    node {
        """
        + DETAIL_NODE_FR+
        """
    }
  }
}
""")


from abc import ABC
from inspect import isawaitable
import logging

logger = logging.getLogger(__name__)

def hookable(hook, overwritable=False):

    def real_decorator(function):
        if isawaitable(function):
            async def wrapped(self, *args, **kwargs):
                if hook in self.hooks.overwritten_hooks:
                    try:
                        assert overwritable, "The function was requested to be overwritten but cannot be overwritten (hookable overwrite false)"
                        return await self.hooks.overwritten_hooks[hook](self, *args, **kwargs)
                    except Exception as e:
                        logger.error(f"Overwritten Hook Exception. Passing {e}")
                        raise e
                        
                if hook in self.hooks.passing_hooks:
                    try:
                        await self.hooks.passing_hooks[hook](self, *args, **kwargs)
                    except Exception as e:
                        logger.error(f"Passing Hook Exception: {e}")

                return await function(self, *args, **kwargs)
        else:
           def wrapped(self, *args, **kwargs):
                if hook in self.hooks.overwritten_hooks:
                    try:
                        assert overwritable, "The function was requested to be overwritten but cannot be overwritten (hookable overwrite false)"
                        return self.hooks.overwritten_hooks[hook](self, *args, **kwargs)
                    except Exception as e:
                        logger.error(f"Overwritten Hook Exception. Passing {e}")
                        raise e
                        
                if hook in self.hooks.passing_hooks:
                    try:
                        self.hooks.passing_hooks[hook](self, *args, **kwargs)
                    except Exception as e:
                        logger.error(f"Passing Hook Exception: {e}")

                return function(self, *args, **kwargs) 


        return wrapped

    return real_decorator



class Hooks:

    def __init__(self) -> None:
        self.overwritten_hooks = {}
        self.passing_hooks = {}
        pass

    def addHook(self, hook, function, overwrite=False):
        if overwrite:
            self.overwritten_hooks[hook] = function
        else:
            self.passing_hooks[hook] = function

class Hookable(ABC):

    def __init__(self, hooks: Hooks=None) -> None:
        super().__init__()
        self.hooks = hooks


import logging

logger = logging.getLogger(__name__)



class TypeMatcher:

    def __init__(self) -> None:
        self.map = {}

    def registerModelForIdentifier(self,identifier, model, overwrite=False):
        identifier = identifier.lower()
        if identifier in self.map: 
            assert overwrite is True, "Attempting to overwrite, please provide overwrite in Call"
            logger.info(f"Overwriting {self.map[identifier]} with {model} for {identifier}")
        self.map[identifier] = model

    def getModelForIdentifier(self, identifier):
        identifier = identifier.lower()
        assert identifier in self.map, f"Nothing stored for for identifier {identifier}, please provide a valid Model and store it"
        return self.map[identifier]




CURRENT_MATCHER = None

def get_current_matcher():
    global CURRENT_MATCHER
    if CURRENT_MATCHER is None:
        CURRENT_MATCHER = TypeMatcher()
   
    return CURRENT_MATCHER


from .provide_critical import ProvideCriticalMessage
from .provide import ProvideMessage
from .provide_done import ProvideDoneMessage
from .provide_progress import ProvideProgressMessage
from .bounced_provide import BouncedProvideMessage


from bergen.schema import Assignation, DataModel, Node, Peasent, Template, Pod, Provision, Transcript, VartPod, Volunteer
from bergen.query import TypedGQL


NEGOTIATION_GQL = TypedGQL("""

  mutation Negotiate(
      $clientType: ClientTypeInput!
      $inward: String,
      $outward: String,
      $port: Int,
      $version: String,
      $pointType: DataPointTypeInput,
      
      ) {
  negotiate(
      clientType: $clientType,
      inward: $inward,
      outward: $outward,
      port: $port,
      version: $version,
      pointType: $pointType
  
  ) {
    timestamp
    extensions
    models {
        identifier
        point {
            id
            outward
            port
            type
        }
    }
    postman {
        type
        kwargs
    }
    provider {
        type
        kwargs
    }
    host {
        type
        kwargs
    }
  }
  }
""", Transcript)


HOST_GQL = TypedGQL("""
    mutation Host($identifier: String!, $extenders: [String]){
        host(identifier: $identifier, extenders: $extenders){
            id
            identifier
            extenders
            point {
                inward
            }
        }
    }
""", DataModel)




from abc import ABC, abstractmethod
from bergen.messages.base import MessageModel
from bergen.debugging import DebugLevel, get_style_for_level
from typing import Generic, TypeVar
from bergen.console import console

T = TypeVar("T")

class Connector(ABC):

    def __init__(self, connection) -> None:
        self.connection = connection

    async def forward(self, message: MessageModel):
        await self.connection.forward(message)




class ContractHandler(Generic[T]):

    def __init__(self, message: T, connection: Connector) -> None:
        self.message = message
        self.connection = connection

    async def log(self, message, level: DebugLevel = DebugLevel.INFO):
        console.log(get_style_for_level(level) + level.value + " : ", message)
    

    async def forward(self, message):
        await self.connection.forward(message)


    


from bergen.debugging import DebugLevel
from bergen.handlers import AssignHandler, ReserveHandler, ProvideHandler
import contextvars


assign_handler_context  = contextvars.ContextVar('assign_handler', default=None)
reserve_handler_context = contextvars.ContextVar("reserve_handler", default=None)
provide_handler_context = contextvars.ContextVar("provide_handler", default=None)


async def log(message, level: DebugLevel = DebugLevel.INFO):
    handler: AssignHandler = assign_handler_context.get()
    await handler.log(message, level)



from bergen.clients.base import BaseBergen

CURRENT_CLIENT = None

def get_current_client() -> BaseBergen:
    global CURRENT_CLIENT
    if CURRENT_CLIENT is None:
        raise Exception("No Client was initialized before")
    else:
        return CURRENT_CLIENT


def set_current_client(arnheim):
    global CURRENT_CLIENT
    CURRENT_CLIENT = arnheim


import asyncio
from abc import ABC, abstractmethod
from asyncio.futures import Future
from bergen.handlers.base import Connector
from bergen.messages import *
from bergen.hookable.base import Hookable
import logging
from typing import Dict, Type
from bergen.models import Pod
from bergen.actors.base import Actor
from functools import partial
from bergen.console import console



logger = logging.getLogger(__name__)



class BaseEntertainer(Hookable):
    ''' Is a mixin for Our Bergen '''
    def __init__(self, raise_exceptions_local=False, client = None, loop=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.provisions = {}
        self.raise_exceptions_local = raise_exceptions_local
        self.loop = loop or asyncio.get_event_loop()
        self.client = client

        self.connector = Connector(self)

        self.template_id_actorClass_map: Dict[str, Type[Actor]] = {}

        self.provision_actor_map: Dict[str, Actor] = {}
        self.provision_actor_run_map: Dict[str, asyncio.Task]= {}
        self.provision_actor_queue_map: Dict[str, asyncio.Queue] = {}


        self.unprovide_futures: Dict[str, Future] = {}
        self.provide_futures: Dict[str, Future] = {}

        self.all_pod_assignments = {}
        self.all_pod_reservations = {}

        self.entertainments = {}
        self.assignments = {}

        self.pending = []

        self.tasks = []

    @abstractmethod
    async def connect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def disconnect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def activateProvision(self):
         raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def deactivateProvision(self):
         raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def forward(self, message: MessageModel):
        raise NotImplementedError("Overwrite this in your Protocol")


    async def on_message(self, message: MessageModel):
        
        if isinstance(message, ProvideDoneMessage): 
            # As we send the Activate Provide request to the Platform we will get a Return Statement
            logger.info("Arkitekt acknowledged the creation of a Pod for our Actor [this has no consequences??]")
            self.provide_futures[message.meta.reference].set_result(message.data.pod)

        if isinstance(message, UnprovideDoneMessage): 
            # As we send the Activate Provide request to the Platform we will get a Return Statement
            logger.info("Arkitekt acknowledged the deletion of the Pod our our former Actor")
            self.unprovide_futures[message.meta.reference].set_result(message.data)

        if isinstance(message, BouncedUnprovideMessage): 
            #  We are responsible To Shut ourselfves down when Arkitekt Tells us t
            assert message.data.provision is not None, "Received Cancellation with no reference to kill" 
            self.unentertain(message, from_arkitekt=True)

        if isinstance(message, BouncedProvideMessage): 
            #  We are responsible To Shut ourselfves down when Arkitekt Tells us t
            assert message.data.template is not None, "Received Cancellation with no reference to kill" 
            logger.info("Arkitekt demands a New Actor for us")
            self.entertain(message, from_arkitekt=True)

        if isinstance(message, BouncedForwardedReserveMessage):
            logger.info("Arkitekt demands a New Reservation from us")
            assert message.data.provision is not None, "Received Forwared Reservation that had no Provision???"
            assert message.data.provision in self.provision_actor_queue_map, f"Provision not entertained {message.data.provision} {self.provision_actor_queue_map}"
            await self.provision_actor_queue_map[message.data.provision].put(message)
            self.all_pod_reservations[message.meta.reference] = message.data.provision # Run in parallel

        if isinstance(message, BouncedUnreserveMessage):
            if message.data.reservation in self.all_pod_reservations: 
                logger.info("Cancellation for Reservatoion received. Canceling!")
                hosting_provising = self.all_pod_reservations[message.data.reservation]
                await self.provision_actor_queue_map[hosting_provising].put(message)
            else:
                logger.error("Received Cancellation for task that was not in our tasks..")


        if isinstance(message, BouncedForwardedAssignMessage):
            assert message.data.provision is not None, "Received assignation that had no Provision???"
            assert message.data.provision in self.provision_actor_queue_map, f"Provision not entertained {message.data.provision} {self.provision_actor_queue_map}"
            await self.provision_actor_queue_map[message.data.provision].put(message)
            self.all_pod_assignments[message.meta.reference] = message.data.provision # Run in parallel

        if isinstance(message, BouncedUnassignMessage):
            if message.data.assignation in self.all_pod_assignments: 
                logger.info("Cancellation for task received. Canceling!")
                hosting_provising = self.all_pod_assignments[message.data.assignation]
                await self.provision_actor_queue_map[hosting_provising].put(message)
            else:
                logger.error("Received Cancellation for task that was not in our tasks..")


    async def deactivateProvision(self, bounced_unprovide: BouncedUnprovideMessage):
        # Where should we do this?
        future = self.loop.create_future()
        self.unprovide_futures[bounced_unprovide.meta.reference] = future
        await self.forward(bounced_unprovide)
        reference = await future # is just a loopback??
        return reference

    async def activateProvision(self, bounced_provide: BouncedProvideMessage):
        
        #We are forwarding the Provision to Arkitekt and wait for its acknowledgement of the creation of this Pod
        future = self.loop.create_future()
        self.provide_futures[bounced_provide.meta.reference] = future
        await self.forward(bounced_provide)
        id = await future
        pod = await Pod.asyncs.get(id=id)
        return pod

    
    def actor_cancelled(self, actor: Actor, future: Future):
        logger.info("Actor is done! Cancellation or Finished??")
        if future.cancelled():
            logger.info("Future was cancelled everything is cools")


    async def get_actorclass_for_template(self, template_id):
        assert template_id in self.template_id_actorClass_map, "We have no Actor stored in our list"
        return self.template_id_actorClass_map[template_id]


    def registerActor(self, template_id: str, actorClass: Type[Actor]):
        assert template_id not in self.template_id_actorClass_map, "We cannot register two Actors for the same template on the same provider"
        self.template_id_actorClass_map[template_id] = actorClass


    async def unentertain(self, bounced_unprovide: BouncedUnprovideMessage, from_arkitekt=False):
        if from_arkitekt: logger.info("Cancellation invoked by Arkitekt")

        provision_reference = bounced_unprovide.data.provision
        deactivation_task = asyncio.create_task(self.deactivateProvision(bounced_unprovide))
        deactivation_task.add_done_callback(print)   

        task = self.provision_actor_run_map[provision_reference]

        if not task.done():
            logger.info("Cancelling Task")
            task.cancel()


        
        
        # THIS Comes form the Arkitekt Platform

    
    async def entertain(self, bounced_provide: BouncedProvideMessage, actorClass: Type[Actor], from_arkitekt=False):
        ''' Takes an instance of a pod, asks arnheim to activate it and accepts requests on it,
        cancel this task to unprovide your local implementatoin '''
        if from_arkitekt: logger.info("Entertainment invoked by Arkitekt")

        provision_reference = bounced_provide.meta.reference # We register Actors unter its provision
        actor = actorClass(self.connector)

        try:
            # Activate Provison , will return
            activation_task = asyncio.create_task(self.activateProvision(bounced_provide))
            activation_task.add_done_callback(print)

            self.provision_actor_map[provision_reference] = actor
            self.provision_actor_queue_map[provision_reference] = actor.queue
            run_task = asyncio.create_task(actor.run(bounced_provide))
            run_task.add_done_callback(partial(self.actor_cancelled, actor))
            self.provision_actor_run_map[provision_reference] = run_task

        except Exception as e:
            logger.error("We have a provision error")
            logger.error(e)
            # This error gets passed back to the provider
            raise e







from enum import Enum
from ..base import MessageDataModel


class ProgressLevel(str, Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"
    ERROR = "ERROR"
    WARNING = "WARN"


class ProgressDataModel(MessageDataModel):
    level: ProgressLevel
    message: str